@@ -8,6 +8,7 @@ import (
 	"encoding/hex"
 	"errors"
 	"fmt"
+	"io"
 	"io/ioutil"
 	"net/url"
 	"os"
@@ -84,19 +85,116 @@ type CommitSummary struct {
 	Subject        string
 }
 
+// Prepend Git config instructions to disable Git LFS filter
+func gitConfigNoLFS(args ...string) []string {
+	// Before git 2.8, setting filters to blank causes lots of warnings, so use cat instead (slightly slower)
+	// Also pre 2.2 it failed completely. We used to use it anyway in git 2.2-2.7 and
+	// suppress the messages in stderr, but doing that with standard StderrPipe suppresses
+	// the git clone output (git thinks it's not a terminal) and makes it look like it's
+	// not working. You can get around that with https://github.com/kr/pty but that
+	// causes difficult issues with passing through Stdin for login prompts
+	// This way is simpler & more practical.
+	filterOverride := ""
+	if !Config.IsGitVersionAtLeast("2.8.0") {
+		filterOverride = "cat"
+	}
+
+	return append([]string{
+		"-c", fmt.Sprintf("filter.lfs.smudge=%v", filterOverride),
+		"-c", fmt.Sprintf("filter.lfs.clean=%v", filterOverride),
+		"-c", "filter.lfs.process=",
+		"-c", "filter.lfs.required=false",
+	}, args...)
+}
+
+// Invoke Git with disabled LFS filters
+func gitNoLFS(args ...string) *subprocess.Cmd {
+	return subprocess.ExecCommand("git", gitConfigNoLFS(args...)...)
+}
+
+func gitNoLFSSimple(args ...string) (string, error) {
+	return subprocess.SimpleExec("git", gitConfigNoLFS(args...)...)
+}
+
+func gitNoLFSBuffered(args ...string) (*subprocess.BufferedCmd, error) {
+	return subprocess.BufferedExec("git", gitConfigNoLFS(args...)...)
+}
+
+// Invoke Git with enabled LFS filters
+func git(args ...string) *subprocess.Cmd {
+	return subprocess.ExecCommand("git", args...)
+}
+
+func gitSimple(args ...string) (string, error) {
+	return subprocess.SimpleExec("git", args...)
+}
+
+func gitBuffered(args ...string) (*subprocess.BufferedCmd, error) {
+	return subprocess.BufferedExec("git", args...)
+}
+
+func CatFile() (*subprocess.BufferedCmd, error) {
+	return gitNoLFSBuffered("cat-file", "--batch-check")
+}
+
+func DiffIndex(ref string, cached bool) (*bufio.Scanner, error) {
+	args := []string{"diff-index", "-M"}
+	if cached {
+		args = append(args, "--cached")
+	}
+	args = append(args, ref)
+
+	cmd, err := gitBuffered(args...)
+	if err != nil {
+		return nil, err
+	}
+	if err = cmd.Stdin.Close(); err != nil {
+		return nil, err
+	}
+
+	return bufio.NewScanner(cmd.Stdout), nil
+}
+
+func HashObject(r io.Reader) (string, error) {
+	cmd := gitNoLFS("hash-object", "--stdin")
+	cmd.Stdin = r
+	out, err := cmd.Output()
+	if err != nil {
+		return "", fmt.Errorf("Error building Git blob OID: %s", err)
+	}
+
+	return string(bytes.TrimSpace(out)), nil
+}
+
+func Log(args ...string) (*subprocess.BufferedCmd, error) {
+	logArgs := append([]string{"log"}, args...)
+	return gitNoLFSBuffered(logArgs...)
+}
+
 func LsRemote(remote, remoteRef string) (string, error) {
 	if remote == "" {
 		return "", errors.New("remote required")
 	}
 	if remoteRef == "" {
-		return subprocess.SimpleExec("git", "ls-remote", remote)
+		return gitNoLFSSimple("ls-remote", remote)
 
 	}
-	return subprocess.SimpleExec("git", "ls-remote", remote, remoteRef)
+	return gitNoLFSSimple("ls-remote", remote, remoteRef)
+}
+
+func LsTree(ref string) (*subprocess.BufferedCmd, error) {
+	return gitNoLFSBuffered(
+		"ls-tree",
+		"-r",          // recurse
+		"-l",          // report object size (we'll need this)
+		"-z",          // null line termination
+		"--full-tree", // start at the root regardless of where we are in it
+		ref,
+	)
 }
 
 func ResolveRef(ref string) (*Ref, error) {
-	outp, err := subprocess.SimpleExec("git", "rev-parse", ref, "--symbolic-full-name", ref)
+	outp, err := gitNoLFSSimple("rev-parse", ref, "--symbolic-full-name", ref)
 	if err != nil {
 		return nil, fmt.Errorf("Git can't resolve ref: %q", ref)
 	}
@@ -199,7 +297,7 @@ func RemoteBranchForLocalBranch(localBranch string) string {
 }
 
 func RemoteList() ([]string, error) {
-	cmd := subprocess.ExecCommand("git", "remote")
+	cmd := gitNoLFS("remote")
 
 	outp, err := cmd.StdoutPipe()
 	if err != nil {
@@ -221,7 +319,7 @@ func RemoteList() ([]string, error) {
 // Refs returns all of the local and remote branches and tags for the current
 // repository. Other refs (HEAD, refs/stash, git notes) are ignored.
 func LocalRefs() ([]*Ref, error) {
-	cmd := subprocess.ExecCommand("git", "show-ref", "--heads", "--tags")
+	cmd := gitNoLFS("show-ref", "--heads", "--tags")
 
 	outp, err := cmd.StdoutPipe()
 	if err != nil {
@@ -277,7 +375,7 @@ func UpdateRefIn(wd string, ref *Ref, to []byte, reason string) error {
 		args = append(args, "-m", reason)
 	}
 
-	cmd := subprocess.ExecCommand("git", args...)
+	cmd := gitNoLFS(args...)
 	cmd.Dir = wd
 
 	return cmd.Run()
@@ -360,9 +458,8 @@ func DefaultRemote() (string, error) {
 	return "", errors.New("Unable to pick default remote, too ambiguous")
 }
 
-func UpdateIndex(file string) error {
-	_, err := subprocess.SimpleExec("git", "update-index", "-q", "--refresh", file)
-	return err
+func UpdateIndexFromStdin() *subprocess.Cmd {
+	return git("update-index", "-q", "--refresh", "--stdin")
 }
 
 type gitConfig struct {
@@ -374,61 +471,61 @@ var Config = &gitConfig{}
 
 // Find returns the git config value for the key
 func (c *gitConfig) Find(val string) string {
-	output, _ := subprocess.SimpleExec("git", "config", val)
+	output, _ := gitSimple("config", val)
 	return output
 }
 
 // FindGlobal returns the git config value global scope for the key
 func (c *gitConfig) FindGlobal(val string) string {
-	output, _ := subprocess.SimpleExec("git", "config", "--global", val)
+	output, _ := gitSimple("config", "--global", val)
 	return output
 }
 
 // FindSystem returns the git config value in system scope for the key
 func (c *gitConfig) FindSystem(val string) string {
-	output, _ := subprocess.SimpleExec("git", "config", "--system", val)
+	output, _ := gitSimple("config", "--system", val)
 	return output
 }
 
 // Find returns the git config value for the key
 func (c *gitConfig) FindLocal(val string) string {
-	output, _ := subprocess.SimpleExec("git", "config", "--local", val)
+	output, _ := gitSimple("config", "--local", val)
 	return output
 }
 
 // SetGlobal sets the git config value for the key in the global config
 func (c *gitConfig) SetGlobal(key, val string) (string, error) {
-	return subprocess.SimpleExec("git", "config", "--global", key, val)
+	return gitSimple("config", "--global", key, val)
 }
 
 // SetSystem sets the git config value for the key in the system config
 func (c *gitConfig) SetSystem(key, val string) (string, error) {
-	return subprocess.SimpleExec("git", "config", "--system", key, val)
+	return gitSimple("config", "--system", key, val)
 }
 
 // UnsetGlobal removes the git config value for the key from the global config
 func (c *gitConfig) UnsetGlobal(key string) (string, error) {
-	return subprocess.SimpleExec("git", "config", "--global", "--unset", key)
+	return gitSimple("config", "--global", "--unset", key)
 }
 
 // UnsetSystem removes the git config value for the key from the system config
 func (c *gitConfig) UnsetSystem(key string) (string, error) {
-	return subprocess.SimpleExec("git", "config", "--system", "--unset", key)
+	return gitSimple("config", "--system", "--unset", key)
 }
 
 // UnsetGlobalSection removes the entire named section from the global config
 func (c *gitConfig) UnsetGlobalSection(key string) (string, error) {
-	return subprocess.SimpleExec("git", "config", "--global", "--remove-section", key)
+	return gitSimple("config", "--global", "--remove-section", key)
 }
 
 // UnsetSystemSection removes the entire named section from the system config
 func (c *gitConfig) UnsetSystemSection(key string) (string, error) {
-	return subprocess.SimpleExec("git", "config", "--system", "--remove-section", key)
+	return gitSimple("config", "--system", "--remove-section", key)
 }
 
 // UnsetLocalSection removes the entire named section from the system config
 func (c *gitConfig) UnsetLocalSection(key string) (string, error) {
-	return subprocess.SimpleExec("git", "config", "--local", "--remove-section", key)
+	return gitSimple("config", "--local", "--remove-section", key)
 }
 
 // SetLocal sets the git config value for the key in the specified config file
@@ -439,7 +536,7 @@ func (c *gitConfig) SetLocal(file, key, val string) (string, error) {
 		args = append(args, "--file", file)
 	}
 	args = append(args, key, val)
-	return subprocess.SimpleExec("git", args...)
+	return gitSimple(args...)
 }
 
 // UnsetLocalKey removes the git config value for the key from the specified config file
@@ -450,17 +547,17 @@ func (c *gitConfig) UnsetLocalKey(file, key string) (string, error) {
 		args = append(args, "--file", file)
 	}
 	args = append(args, "--unset", key)
-	return subprocess.SimpleExec("git", args...)
+	return gitSimple(args...)
 }
 
 // List lists all of the git config values
 func (c *gitConfig) List() (string, error) {
-	return subprocess.SimpleExec("git", "config", "-l")
+	return gitSimple("config", "-l")
 }
 
 // ListFromFile lists all of the git config values in the given config file
 func (c *gitConfig) ListFromFile(f string) (string, error) {
-	return subprocess.SimpleExec("git", "config", "-l", "-f", f)
+	return gitSimple("config", "-l", "-f", f)
 }
 
 // Version returns the git version
@@ -469,7 +566,7 @@ func (c *gitConfig) Version() (string, error) {
 	defer c.mu.Unlock()
 
 	if len(c.gitVersion) == 0 {
-		v, err := subprocess.SimpleExec("git", "version")
+		v, err := gitSimple("version")
 		if err != nil {
 			return v, err
 		}
@@ -496,7 +593,7 @@ func (c *gitConfig) IsGitVersionAtLeast(ver string) bool {
 // includeRemoteBranches: true to include refs on remote branches
 // onlyRemote: set to non-blank to only include remote branches on a single remote
 func RecentBranches(since time.Time, includeRemoteBranches bool, onlyRemote string) ([]*Ref, error) {
-	cmd := subprocess.ExecCommand("git", "for-each-ref",
+	cmd := gitNoLFS("for-each-ref",
 		`--sort=-committerdate`,
 		`--format=%(refname) %(objectname) %(committerdate:iso)`,
 		"refs")
@@ -599,7 +696,7 @@ func FormatGitDate(tm time.Time) string {
 
 // Get summary information about a commit
 func GetCommitSummary(commit string) (*CommitSummary, error) {
-	cmd := subprocess.ExecCommand("git", "show", "-s",
+	cmd := gitNoLFS("show", "-s",
 		`--format=%H|%h|%P|%ai|%ci|%ae|%an|%ce|%cn|%s`, commit)
 
 	out, err := cmd.CombinedOutput()
@@ -634,7 +731,7 @@ func GetCommitSummary(commit string) (*CommitSummary, error) {
 }
 
 func GitAndRootDirs() (string, string, error) {
-	cmd := subprocess.ExecCommand("git", "rev-parse", "--git-dir", "--show-toplevel")
+	cmd := gitNoLFS("rev-parse", "--git-dir", "--show-toplevel")
 	buf := &bytes.Buffer{}
 	cmd.Stderr = buf
 
@@ -669,7 +766,7 @@ func GitAndRootDirs() (string, string, error) {
 }
 
 func RootDir() (string, error) {
-	cmd := subprocess.ExecCommand("git", "rev-parse", "--show-toplevel")
+	cmd := gitNoLFS("rev-parse", "--show-toplevel")
 	out, err := cmd.Output()
 	if err != nil {
 		return "", fmt.Errorf("Failed to call git rev-parse --show-toplevel: %v %v", err, string(out))
@@ -685,7 +782,7 @@ func RootDir() (string, error) {
 }
 
 func GitDir() (string, error) {
-	cmd := subprocess.ExecCommand("git", "rev-parse", "--git-dir")
+	cmd := gitNoLFS("rev-parse", "--git-dir")
 	out, err := cmd.Output()
 	if err != nil {
 		return "", fmt.Errorf("Failed to call git rev-parse --git-dir: %v %v", err, string(out))
@@ -886,25 +983,7 @@ type CloneFlags struct {
 // so that files in the working copy will be pointers and not real LFS data
 func CloneWithoutFilters(flags CloneFlags, args []string) error {
 
-	// Before git 2.8, setting filters to blank causes lots of warnings, so use cat instead (slightly slower)
-	// Also pre 2.2 it failed completely. We used to use it anyway in git 2.2-2.7 and
-	// suppress the messages in stderr, but doing that with standard StderrPipe suppresses
-	// the git clone output (git thinks it's not a terminal) and makes it look like it's
-	// not working. You can get around that with https://github.com/kr/pty but that
-	// causes difficult issues with passing through Stdin for login prompts
-	// This way is simpler & more practical.
-	filterOverride := ""
-	if !Config.IsGitVersionAtLeast("2.8.0") {
-		filterOverride = "cat"
-	}
-	// Disable the LFS filters while cloning to speed things up
-	// this is especially effective on Windows where even calling git-lfs at all
-	// with --skip-smudge is costly across many files in a checkout
-	cmdargs := []string{
-		"-c", fmt.Sprintf("filter.lfs.smudge=%v", filterOverride),
-		"-c", "filter.lfs.process=",
-		"-c", "filter.lfs.required=false",
-		"clone"}
+	cmdargs := []string{"clone"}
 
 	// flags
 	if flags.Bare {
@@ -1000,7 +1079,7 @@ func CloneWithoutFilters(flags CloneFlags, args []string) error {
 
 	// Now args
 	cmdargs = append(cmdargs, args...)
-	cmd := subprocess.ExecCommand("git", cmdargs...)
+	cmd := gitNoLFS(cmdargs...)
 
 	// Assign all streams direct
 	cmd.Stdout = os.Stdout
@@ -1039,7 +1118,7 @@ func Checkout(treeish string, paths []string, force bool) error {
 		args = append(args, append([]string{"--"}, paths...)...)
 	}
 
-	_, err := subprocess.SimpleExec("git", args...)
+	_, err := gitNoLFSSimple(args...)
 	return err
 }
 
@@ -1047,7 +1126,7 @@ func Checkout(treeish string, paths []string, force bool) error {
 // currently cached locally. No remote request is made to verify them.
 func CachedRemoteRefs(remoteName string) ([]*Ref, error) {
 	var ret []*Ref
-	cmd := subprocess.ExecCommand("git", "show-ref")
+	cmd := gitNoLFS("show-ref")
 
 	outp, err := cmd.StdoutPipe()
 	if err != nil {
@@ -1072,11 +1151,21 @@ func CachedRemoteRefs(remoteName string) ([]*Ref, error) {
 	return ret, cmd.Wait()
 }
 
+// Fetch performs a fetch with no arguments against the given remotes.
+func Fetch(remotes ...string) error {
+	if len(remotes) == 0 {
+		return nil
+	}
+
+	_, err := gitNoLFSSimple(append([]string{"fetch"}, remotes...)...)
+	return err
+}
+
 // RemoteRefs returns a list of branches & tags for a remote by actually
 // accessing the remote vir git ls-remote
 func RemoteRefs(remoteName string) ([]*Ref, error) {
 	var ret []*Ref
-	cmd := subprocess.ExecCommand("git", "ls-remote", "--heads", "--tags", "-q", remoteName)
+	cmd := gitNoLFS("ls-remote", "--heads", "--tags", "-q", remoteName)
 
 	outp, err := cmd.StdoutPipe()
 	if err != nil {
@@ -1115,7 +1204,7 @@ func AllRefs() ([]*Ref, error) {
 // the given working directory "wd", or an error if those references could not
 // be loaded.
 func AllRefsIn(wd string) ([]*Ref, error) {
-	cmd := subprocess.ExecCommand("git",
+	cmd := gitNoLFS(
 		"for-each-ref", "--format=%(objectname)%00%(refname)")
 	cmd.Dir = wd
 
@@ -1161,7 +1250,7 @@ func GetTrackedFiles(pattern string) ([]string, error) {
 	rootWildcard := len(safePattern) < len(pattern) && strings.ContainsRune(safePattern, '*')
 
 	var ret []string
-	cmd := subprocess.ExecCommand("git",
+	cmd := gitNoLFS(
 		"-c", "core.quotepath=false", // handle special chars in filenames
 		"ls-files",
 		"--cached", // include things which are staged but not committed right now
@@ -1220,7 +1309,7 @@ func GetFilesChanged(from, to string) ([]string, error) {
 	}
 	args = append(args, "--") // no ambiguous patterns
 
-	cmd := subprocess.ExecCommand("git", args...)
+	cmd := gitNoLFS(args...)
 	outp, err := cmd.StdoutPipe()
 	if err != nil {
 		return nil, fmt.Errorf("Failed to call git diff: %v", err)
@@ -1251,7 +1340,7 @@ func IsFileModified(filepath string) (bool, error) {
 		"--", // separator in case filename ambiguous
 		filepath,
 	}
-	cmd := subprocess.ExecCommand("git", args...)
+	cmd := git(args...)
 	outp, err := cmd.StdoutPipe()
 	if err != nil {
 		return false, lfserrors.Wrap(err, "Failed to call git status")
